/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include <mysql.h>

#ifdef TIME_WITH_SYS_TIME
#   include <sys/time.h>
#   include <time.h>
#else
#   ifdef HAVE_SYS_TIME_H
#       include <sys/time.h>
#   else
#       include <time.h>
#   endif
#endif

#include "storage_driver.h"
#include "mysql_drv.h"
#include "libdspam.h"
#include "config.h"
#include "error.h"
#include "language.h"
#include "util.h"
#include "lht.h"

int test(MYSQL *, char *);

int test(MYSQL *dbh, char *query) {
  int i;
//  printf("called: %s\n", query);
  i = mysql_query(dbh, query);
//  if (i) { printf("called: %s\n", mysql_error(dbh)); }
  return i;
}

#define MYSQL_RUN_QUERY(A, B) test(A, B)


int
dspam_init_driver (void)
{
  return 0;
}

int
dspam_shutdown_driver (void)
{
  return 0;
}

int
_mysql_drv_get_spamtotals (DSPAM_CTX * CTX)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  char query[1024];
  struct passwd *p;
  MYSQL_RES *result;
  MYSQL_ROW row;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_mysql_drv_get_spamtotals: invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_mysql_drv_get_spamtotals: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  snprintf (query, sizeof (query),
            "select total_spam, total_innocent, spam_misses, false_positives, spam_corpusfed, innocent_corpusfed from dspam_stats where uid = %d",
            p->pw_uid);
  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return EFAILURE;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL)
    return EFAILURE;

  row = mysql_fetch_row (result);
  if (row == NULL)
  {
    mysql_free_result (result);
    return EFAILURE;
  }

  CTX->totals.total_spam = strtol (row[0], NULL, 0);
  CTX->totals.total_innocent = strtol (row[1], NULL, 0);
  CTX->totals.spam_misclassified = strtol (row[2], NULL, 0);
  CTX->totals.innocent_misclassified = strtol (row[3], NULL, 0);
  CTX->totals.spam_corpusfed = strtol (row[4], NULL, 0);
  CTX->totals.innocent_corpusfed = strtol (row[5], NULL, 0);

  mysql_free_result (result);

  return 0;
}

int
_mysql_drv_set_spamtotals (DSPAM_CTX * CTX)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  char query[1024];
  int result = 0;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_mysql_drv_set_spamtotals: invalid database handle (NULL)");
    return EINVAL;
  }

  if (CTX->mode == DSM_CLASSIFY)
  {
    _mysql_drv_get_spamtotals (CTX);    /* undo changes to in memory totals */
    return 0;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_mysql_drv_get_spamtotals: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  if (s->control_totals.total_innocent == 0)
  {
    snprintf (query, sizeof (query),
              "insert into dspam_stats(uid, total_spam, total_innocent, spam_misses, false_positives, spam_corpusfed, innocent_corpusfed) values(%d, %ld, %ld, %ld, %ld, %ld, %ld)",
              p->pw_uid, CTX->totals.total_spam,
              CTX->totals.total_innocent, CTX->totals.spam_misclassified,
              CTX->totals.innocent_misclassified, CTX->totals.spam_corpusfed,
              CTX->totals.innocent_corpusfed);
    result = MYSQL_RUN_QUERY (s->dbh, query);
  }

  if (s->control_totals.total_innocent != 0 || result)
  {
    snprintf (query, sizeof (query),
              "update dspam_stats set total_spam = total_spam %s %d, "
              "total_innocent = total_innocent %s %d, "
              "spam_misses = spam_misses %s %d, "
              "false_positives = false_positives %s %d, "
              "spam_corpusfed = spam_corpusfed %s %d, "
              "innocent_corpusfed = innocent_corpusfed %s %d "
              "where uid = %d",
              (CTX->totals.total_spam >
               s->control_totals.total_spam) ? "+" : "-",
              abs (CTX->totals.total_spam -
                   s->control_totals.total_spam),
              (CTX->totals.total_innocent >
               s->control_totals.total_innocent) ? "+" : "-",
              abs (CTX->totals.total_innocent -
                   s->control_totals.total_innocent),
              (CTX->totals.spam_misclassified >
               s->control_totals.spam_misclassified) ? "+" : "-",
              abs (CTX->totals.spam_misclassified -
                   s->control_totals.spam_misclassified),
              (CTX->totals.innocent_misclassified >
               s->control_totals.innocent_misclassified) ? "+" : "-",
              abs (CTX->totals.innocent_misclassified -
                   s->control_totals.innocent_misclassified),
              (CTX->totals.spam_corpusfed >
               s->control_totals.spam_corpusfed) ? "+" : "-",
              abs (CTX->totals.spam_corpusfed -
                   s->control_totals.spam_corpusfed),
              (CTX->totals.innocent_corpusfed >
               s->control_totals.innocent_corpusfed) ? "+" : "-",
              abs (CTX->totals.innocent_corpusfed -
                   s->control_totals.innocent_corpusfed), p->pw_uid);

    if (MYSQL_RUN_QUERY (s->dbh, query))
    {
      _mysql_drv_query_error (mysql_error (s->dbh), query);
      return EFAILURE;
    }
  }

  return 0;
}

int
_ds_getall_spamrecords (DSPAM_CTX * CTX, struct lht *freq)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  buffer *query;
  struct lht_node *node_lht;
  struct lht_c c_lht;
  char scratch[1024];
  MYSQL_RES *result;
  MYSQL_ROW row;
  struct _ds_spam_stat stat;
  unsigned long long token = 0;
  int get_one = 0;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_getall_spamrecords: invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_getall_spamrecords: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  stat.spam_hits = 0;
  stat.innocent_hits = 0;

  query = buffer_create (NULL);
  if (query == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  snprintf (scratch, sizeof (scratch),
            "select token, spam_hits, innocent_hits from dspam_token_data where uid = %d and token in(",
            p->pw_uid);
  buffer_cat (query, scratch);
  node_lht = c_lht_first (freq, &c_lht);
  while (node_lht != NULL)
  {
    snprintf (scratch, sizeof (scratch), "'%llu'", node_lht->key);
    buffer_cat (query, scratch);
    node_lht->s.innocent_hits = 0;
    node_lht->s.spam_hits = 0;
    node_lht->s.probability = 0;
    node_lht->s.disk = 'N';
    node_lht = c_lht_next (freq, &c_lht);
    if (node_lht != NULL)
      buffer_cat (query, ",");
    get_one = 1;
  }
  buffer_cat (query, ")");

#ifdef VERBOSE
  LOGDEBUG ("mysql query length: %ld\n", query->used);
  _mysql_drv_query_error ("verbose debug", query->data);
#endif

  if (!get_one)
    return EINVAL;

  if (MYSQL_RUN_QUERY (s->dbh, query->data))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query->data);
    return EFAILURE;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL)
    return EFAILURE;

  stat.probability = 0;
  while ((row = mysql_fetch_row (result)) != NULL)
  {
    token = strtoull (row[0], NULL, 0);
    stat.spam_hits = strtol (row[1], NULL, 0);
    stat.innocent_hits = strtol (row[2], NULL, 0);
    stat.disk = 'Y';

    /* our control token will tell us how each token was altered */
    if (stat.spam_hits && stat.innocent_hits && !s->control_token)
    {
      s->control_token = token;
      s->control_sh = stat.spam_hits;
      s->control_ih = stat.innocent_hits;
    }
    lht_setspamstat (freq, token, &stat);
  }

  if (!s->control_token)
  {
    s->control_token = token;
    s->control_sh = stat.spam_hits;
    s->control_ih = stat.innocent_hits;
  }

  mysql_free_result (result);
  buffer_destroy (query);
  return 0;
}

int
_ds_setall_spamrecords (DSPAM_CTX * CTX, struct lht *freq)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct _ds_spam_stat stat;
  struct lht_node *node_lht;
  struct lht_c c_lht;
  buffer *query;
  char scratch[1024];
  struct passwd *p;
  int update_one = 0;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_setall_spamrecords: invalid database handle (NULL)");
    return EINVAL;
  }

  if (CTX->mode == DSM_CLASSIFY)
    return 0;

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_setall_spamrecords: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  query = buffer_create (NULL);
  if (query == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  if (s->control_token == 0)
  {
    node_lht = c_lht_first (freq, &c_lht);
    if (node_lht == NULL)
    {
      stat.spam_hits = 0;
      stat.innocent_hits = 0;
    }
    else
    {
      stat.spam_hits = node_lht->s.spam_hits;
      stat.innocent_hits = node_lht->s.innocent_hits;
    }
  }
  else
  {
    lht_getspamstat (freq, s->control_token, &stat);
  }

  snprintf (scratch, sizeof (scratch),
            "update dspam_token_data set last_hit = current_date(), "
            "spam_hits = spam_hits %s %d, "
            "innocent_hits = innocent_hits %s %d "
            "where uid = %d and token in(",
            (stat.spam_hits > s->control_sh) ? "+" : "-",
            abs (stat.spam_hits - s->control_sh),
            (stat.innocent_hits > s->control_ih) ? "+" : "-",
            abs (stat.innocent_hits - s->control_ih), p->pw_uid);

  buffer_cat (query, scratch);

  node_lht = c_lht_first (freq, &c_lht);
  while (node_lht != NULL)
  {
    int wrote_this = 0;
    lht_getspamstat (freq, node_lht->key, &stat);
    if (stat.disk != 'Y')
    {
      char insert[1024];
      if (query->used && query->data[strlen (query->data) - 1] == ',')
      {
        query->used--;
        query->data[strlen (query->data) - 1] = 0;
      }
      snprintf (insert, sizeof (insert),
                "insert into dspam_token_data(uid, token, spam_hits, innocent_hits, last_hit) values(%d, '%llu', %ld, %ld, current_date())",
                p->pw_uid, node_lht->key, stat.spam_hits, stat.innocent_hits);
      if (MYSQL_RUN_QUERY (s->dbh, insert))
      {
        /* possible that it got inserted in-between dspam agent runs? */
#ifdef DEBUG
        _mysql_drv_query_error (mysql_error (s->dbh), insert);
#endif
        stat.disk = 'Y';
      }
    }

    if (stat.disk == 'Y')
    {
      snprintf (scratch, sizeof (scratch), "'%llu'", node_lht->key);
      buffer_cat (query, scratch);
      update_one = 1;
      wrote_this = 1;
      node_lht->s.disk = 'Y';
    }
    node_lht = c_lht_next (freq, &c_lht);
    if (node_lht != NULL && wrote_this)
      buffer_cat (query, ",");
  }
  buffer_cat (query, ")");
#ifdef TUM
  if (CTX->mode == DSM_PROCESS && ! (CTX->flags & DSF_CORPUS))
    buffer_cat (query, " and innocent_hits + spam_hits < 25");
#endif

  if (stat.spam_hits < s->control_sh)
    buffer_cat (query,  " and spam_hits > 0");
                                                                                
  if (stat.innocent_hits < s->control_ih)
    buffer_cat (query, " and innocent_hits > 0");

  if (update_one)
  {
    if (MYSQL_RUN_QUERY (s->dbh, query->data))
    {
      _mysql_drv_query_error (mysql_error (s->dbh), query->data);
      return EFAILURE;
    }
  }

  buffer_destroy (query);
  return 0;
}

int
_ds_get_spamrecord (DSPAM_CTX * CTX, unsigned long long token,
                    struct _ds_spam_stat *stat)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  char query[1024];
  struct passwd *p;
  MYSQL_RES *result;
  MYSQL_ROW row;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_get_spamrecord: invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_get_spamrecord: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  snprintf (query, sizeof (query),
            "select spam_hits, innocent_hits from dspam_token_data "
            "where uid = %d " "and token = '%llu' ", p->pw_uid, token);

  stat->probability = 0.0;
  stat->spam_hits = 0;
  stat->innocent_hits = 0;
  stat->disk = 'N';

  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return EFAILURE;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL)
    return EFAILURE;

  row = mysql_fetch_row (result);
  if (row == NULL)
  {
    mysql_free_result (result);
    return 0;
  }

  stat->spam_hits = strtol (row[0], NULL, 0);
  stat->innocent_hits = strtol (row[1], NULL, 0);
  stat->disk = 'Y';
  mysql_free_result (result);
  return 0;
}

int
_ds_set_spamrecord (DSPAM_CTX * CTX, unsigned long long token,
                    struct _ds_spam_stat *stat)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  char query[1024];
  struct passwd *p;
  int result = 0;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_set_spamrecord: invalid database handle (NULL)");
    return EINVAL;
  }

  if (CTX->mode == DSM_CLASSIFY)
    return 0;

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_set_spamrecord: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  /* It's either not on disk or the caller isn't using stat.disk */
  if (stat->disk != 'Y')
  {
    snprintf (query, sizeof (query),
              "insert into dspam_token_data(uid, token, spam_hits, innocent_hits, last_hit)"
              " values(%d, '%llu', %ld, %ld, current_date())",
              p->pw_uid, token, stat->spam_hits, stat->innocent_hits);
    result = MYSQL_RUN_QUERY (s->dbh, query);
  }

  if (stat->disk == 'Y' || result)
  {
    /* insert failed; try updating instead */
    snprintf (query, sizeof (query), "update dspam_token_data "
              "set spam_hits = %ld, "
              "innocent_hits = %ld "
              "where uid = %d "
              "and token = %lld", stat->spam_hits,
              stat->innocent_hits, p->pw_uid, token);
#ifdef TUM
    if (CTX->mode == DSM_PROCESS && !(CTX->flags & DSF_CORPUS))
      strlcat(query, " and innocent_hits + spam_hits < 25", sizeof(query));
#endif

    if (MYSQL_RUN_QUERY (s->dbh, query))
    {
      _mysql_drv_query_error (mysql_error (s->dbh), query);
      return EFAILURE;
    }
  }

  return 0;
}

int
_ds_init_storage (DSPAM_CTX * CTX)
{
  struct _mysql_drv_storage *s;
  MYSQL *connect;
  FILE *file;
  char filename[MAX_FILENAME_LENGTH];
  char buffer[128];
  char hostname[128];
  char user[64];
  char password[32];
  char db[64];
  int port = 3306, i = 0;

  /* don't init if we're already initted */
  if (CTX->storage != NULL)
  {
    LOGDEBUG ("_ds_init_storage: storage already initialized");
    return EINVAL;
  }

  s = malloc (sizeof (struct _mysql_drv_storage));
  if (s == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  s->dbh = NULL;
  s->control_token = 0;
  s->iter_user = NULL;
  s->iter_token = NULL;
  s->iter_sig = NULL;

  snprintf (filename, MAX_FILENAME_LENGTH, "%s/mysql.data", USERDIR);
  file = fopen (filename, "r");
  if (file == NULL)
  {
    LOG (LOG_WARNING, "unable to open %s for reading: %s",
         filename, strerror (errno));
    return EFAILURE;
  }

  db[0] = 0;

  while (fgets (buffer, sizeof (buffer), file) != NULL)
  {
    chomp (buffer);
    if (!i)
      strlcpy (hostname, buffer, sizeof (hostname));
    else if (i == 1)
      port = atoi (buffer);
    else if (i == 2)
      strlcpy (user, buffer, sizeof (user));
    else if (i == 3)
      strlcpy (password, buffer, sizeof (password));
    else if (i == 4)
      strlcpy (db, buffer, sizeof (db));
    i++;
  }
  fclose (file);

  if (db[0] == 0)
  {
    LOG (LOG_WARNING, "file %s: incomplete mysql connect data", filename);
    return EINVAL;
  }

  s->dbh = mysql_init (NULL);
  if (s->dbh == NULL)
  {
    LOGDEBUG
      ("_ds_init_storage: mysql_init: unable to initialize handle to database");
    return EUNKNOWN;
  }

  if (hostname[0] == '/')
  {
    connect =
      mysql_real_connect (s->dbh, NULL, user, password, db, 0, hostname, 
                          REAL_CONNECT_FLAG);

  }
  else
  {
    connect =
      mysql_real_connect (s->dbh, hostname, user, password, db, port, NULL,
                          REAL_CONNECT_FLAG);
  }
  if (connect == NULL)
  {
    LOG (LOG_WARNING, "%s", mysql_error (s->dbh));
    return EFAILURE;
  }

  CTX->storage = s;

  /* get spam totals on successful init */
  if (CTX->username != NULL)
  {
      if (_mysql_drv_get_spamtotals (CTX))
      {
        LOGDEBUG ("unable to load totals.  using zero values.");
        memset (&CTX->totals, 0, sizeof (struct _ds_spam_totals));
        memset (&s->control_totals, 0, sizeof (struct _ds_spam_totals));
      }
      else
      {
        memcpy (&s->control_totals, &CTX->totals,
                sizeof (struct _ds_spam_totals));
      }
  }
  else
  {
    memset (&CTX->totals, 0, sizeof (struct _ds_spam_totals));
    memset (&s->control_totals, 0, sizeof (struct _ds_spam_totals));
  }

  return 0;
}

int
_ds_shutdown_storage (DSPAM_CTX * CTX)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_shutdown_storage: invalid database handle (NULL)");
    return EINVAL;
  }

  /* Store spam totals on shutdown */
  if (CTX->username != NULL && CTX->mode != DSM_CLASSIFY)
  {
      _mysql_drv_set_spamtotals (CTX);
  }

  mysql_close (s->dbh);
  s->dbh = NULL;

  return 0;
}

int
_ds_create_signature_id (DSPAM_CTX * CTX, char *buf, size_t len)
{
  char session[64];
  char digit[6];
  int pid, j;

  pid = getpid ();
  snprintf (session, sizeof (session), "%8lx%d", time (NULL), pid);

  for (j = 0; j < 2; j++)
  {
    snprintf (digit, 6, "%d", rand ());
    strlcat (session, digit, 64);
  }

  strlcpy (buf, session, len);
  return 0;
}

int
_ds_get_signature (DSPAM_CTX * CTX, struct _ds_spam_signature *SIG,
                   const char *signature)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  unsigned long *lengths;
  char *mem;
  char query[128];
  MYSQL_RES *result;
  MYSQL_ROW row;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_get_signature: invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_get_signature: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  snprintf (query, sizeof (query),
            "select data, length from dspam_signature_data where uid = %d and signature = \"%s\"",
            p->pw_uid, signature);
  if (mysql_real_query (s->dbh, query, strlen (query)))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return EFAILURE;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL)
    return EFAILURE;

  row = mysql_fetch_row (result);
  if (row == NULL)
  {
    mysql_free_result (result);
    return EFAILURE;
  }

  lengths = mysql_fetch_lengths (result);
  if (lengths == NULL || lengths[0] == 0)
  {
    mysql_free_result (result);
    return EFAILURE;
  }

  mem = malloc (lengths[0]);
  if (mem == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    mysql_free_result (result);
    return EUNKNOWN;
  }

  memcpy (mem, row[0], lengths[0]);
  SIG->data = mem;
  SIG->length = strtol (row[1], NULL, 0);

  mysql_free_result (result);

  return 0;
}

int
_ds_set_signature (DSPAM_CTX * CTX, struct _ds_spam_signature *SIG,
                   const char *signature)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  unsigned long length;
  char *mem;
  char scratch[1024];
  buffer *query;
  struct passwd *p;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_set_signature; invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_set_signature: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  query = buffer_create (NULL);
  if (query == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  mem = malloc (SIG->length * 2);
  if (mem == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }
  length = mysql_real_escape_string (s->dbh, mem, SIG->data, SIG->length);

  snprintf (scratch, sizeof (scratch),
            "insert into dspam_signature_data(uid, signature, length, created_on, data) values(%d, \"%s\", %ld, current_date(), \"",
            p->pw_uid, signature, SIG->length);
  buffer_cat (query, scratch);
  buffer_cat (query, mem);
  buffer_cat (query, "\")");

  if (mysql_real_query (s->dbh, query->data, query->used))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query->data);
    return EFAILURE;
  }

  free (mem);
  return 0;
}

int
_ds_delete_signature (DSPAM_CTX * CTX, const char *signature)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  char query[128];

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_delete_signature: invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_delete_signature: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  snprintf (query, sizeof (query),
            "delete from dspam_signature_data where uid = %d and signature = \"%s\"",
            p->pw_uid, signature);
  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return EFAILURE;
  }

  return 0;
}

int
_ds_verify_signature (DSPAM_CTX * CTX, const char *signature)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  char query[128];
  MYSQL_RES *result;
  MYSQL_ROW row;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_verify_signature: invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_verisy_signature: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  snprintf (query, sizeof (query),
            "select signature from dspam_signature_data where uid = %d and signature = \"%s\"",
            p->pw_uid, signature);
  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return EFAILURE;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL)
    return -1;

  row = mysql_fetch_row (result);
  if (row == NULL)
  {
    mysql_free_result (result);
    return -1;
  }

  mysql_free_result (result);
  return 0;
}

char *
_ds_get_nextuser (DSPAM_CTX * CTX)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  static char user[MAX_FILENAME_LENGTH];
#ifndef VIRTUAL_USERS
  struct passwd *p;
  uid_t uid;
#endif
  char query[128];
  MYSQL_ROW row;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_get_nextuser: invalid database handle (NULL)");
    return NULL;
  }

  if (s->iter_user == NULL)
  {
#ifdef VIRTUAL_USERS
    strcpy (query, "select distinct username from dspam_virtual_uids");
#else
    strcpy (query, "select distinct uid from dspam_stats");
#endif
    if (MYSQL_RUN_QUERY (s->dbh, query))
    {
      _mysql_drv_query_error (mysql_error (s->dbh), query);
      return NULL;
    }

    s->iter_user = mysql_use_result (s->dbh);
    if (s->iter_user == NULL)
      return NULL;
  }

  row = mysql_fetch_row (s->iter_user);
  if (row == NULL)
  {
    mysql_free_result (s->iter_user);
    s->iter_user = NULL;
    return NULL;
  }

#ifdef VIRTUAL_USERS
  strlcpy (user, row[0], sizeof (user));
#else
  uid = (uid_t) atoi (row[0]);
  p = _mysql_drv_getpwuid (CTX, uid);
  if (p == NULL)
  {
    mysql_free_result (s->iter_user);
    s->iter_user = NULL;
    return NULL;
  }

  strlcpy (user, p->pw_name, sizeof (user));
#endif

  return user;
}

struct _ds_storage_record *
_ds_get_nexttoken (DSPAM_CTX * CTX)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct _ds_storage_record *st;
  char query[128];
  MYSQL_ROW row;
  struct passwd *p;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_get_nexttoken: invalid database handle (NULL)");
    return NULL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_get_nexttoken: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return NULL;
  }

  st = malloc (sizeof (struct _ds_storage_record));
  if (st == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return NULL;
  }

  memset (st, 0, sizeof (struct _ds_storage_record));

  if (s->iter_token == NULL)
  {
    snprintf (query, sizeof (query),
              "select token, spam_hits, innocent_hits, unix_timestamp(last_hit) from dspam_token_data where uid = %d",
              p->pw_uid);
    if (MYSQL_RUN_QUERY (s->dbh, query))
    {
      _mysql_drv_query_error (mysql_error (s->dbh), query);
      return NULL;
    }

    s->iter_token = mysql_use_result (s->dbh);
    if (s->iter_token == NULL)
      return NULL;
  }

  row = mysql_fetch_row (s->iter_token);
  if (row == NULL)
  {
    mysql_free_result (s->iter_token);
    s->iter_token = NULL;
    return NULL;
  }

  st->token = strtoull (row[0], NULL, 0);
  st->spam_hits = strtol (row[1], NULL, 0);
  st->innocent_hits = strtol (row[2], NULL, 0);
  st->last_hit = (time_t) strtol (row[3], NULL, 0);

  return st;
}

struct _ds_storage_signature *
_ds_get_nextsignature (DSPAM_CTX * CTX)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct _ds_storage_signature *st;
  unsigned long *lengths;
  char query[128];
  MYSQL_ROW row;
  struct passwd *p;
  char *mem;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_get_nextsignature: invalid database handle (NULL)");
    return NULL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_get_nextsignature: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return NULL;
  }

  st = malloc (sizeof (struct _ds_storage_signature));
  if (st == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return NULL;
  }

  memset (st, 0, sizeof (struct _ds_storage_signature));

  if (s->iter_sig == NULL)
  {
    snprintf (query, sizeof (query),
              "select data, signature, length, unix_timestamp(created_on) from dspam_signature_data where uid = %d",
              p->pw_uid);
    if (mysql_real_query (s->dbh, query, strlen (query)))
    {
      _mysql_drv_query_error (mysql_error (s->dbh), query);
      return NULL;
    }

    s->iter_sig = mysql_use_result (s->dbh);
    if (s->iter_sig == NULL)
      return NULL;
  }

  row = mysql_fetch_row (s->iter_sig);
  if (row == NULL)
  {
    mysql_free_result (s->iter_sig);
    s->iter_sig = NULL;
    return NULL;
  }

  lengths = mysql_fetch_lengths (s->iter_sig);
  if (lengths == NULL || lengths[0] == 0)
  {
    mysql_free_result (s->iter_sig);
    return NULL;
  }

  mem = malloc (lengths[0]);
  if (mem == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    mysql_free_result (s->iter_sig);
    return NULL;
  }

  memcpy (mem, row[0], lengths[0]);
  st->data = mem;
  strlcpy (st->signature, row[1], sizeof (st->signature));
  st->length = strtol (row[2], NULL, 0);
  st->created_on = (time_t) strtol (row[3], NULL, 0);

  return st;
}

struct passwd *
_mysql_drv_getpwnam (DSPAM_CTX * CTX, const char *name)
{
#ifndef VIRTUAL_USERS
  static struct passwd p = { NULL, NULL, 0, 0, NULL, NULL, NULL };
  struct passwd *q;

  if (p.pw_name != NULL)
  {
    /* cache the last name queried */
    if (name != NULL && !strcmp (p.pw_name, name))
      return &p;

    free (p.pw_name);
    p.pw_name = NULL;
    p.pw_uid = 0;
  }

  q = getpwnam (name);
  if (q == NULL)
    return NULL;
  p.pw_uid = q->pw_uid;
  p.pw_name = strdup (q->pw_name);

  return &p;
#else
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  static struct passwd p = { NULL, NULL, 0, 0, NULL, NULL, NULL };
  char query[256];
  MYSQL_RES *result;
  MYSQL_ROW row;

  if (p.pw_name != NULL)
  {
    /* cache the last name queried */
    if (name != NULL && !strcmp (p.pw_name, name))
      return &p;

    free (p.pw_name);
    p.pw_name = NULL;
  }

  snprintf (query, sizeof (query),
            "select uid from dspam_virtual_uids where username = '%s'", name);

  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return NULL;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL) {
    return _mysql_drv_setpwnam (CTX, name);
  }

  row = mysql_fetch_row (result);
  if (row == NULL)
  {
    mysql_free_result (result);
    return _mysql_drv_setpwnam (CTX, name);
  }

  if (row[0] == NULL)
  {
    mysql_free_result (result);
    return _mysql_drv_setpwnam (CTX, name);
  }

  p.pw_uid = strtol (row[0], NULL, 0);
  p.pw_name = strdup (name);

  mysql_free_result (result);
  return &p;

#endif
}

struct passwd *
_mysql_drv_getpwuid (DSPAM_CTX * CTX, uid_t uid)
{
#ifndef VIRTUAL_USERS
  static struct passwd p = { NULL, NULL, 0, 0, NULL, NULL, NULL };
  struct passwd *q;

  if (p.pw_name != NULL)
  {
    /* cache the last uid queried */
    if (p.pw_uid == uid)
    {
      return &p;
    }
    p.pw_name = NULL;
  }

  q = getpwuid (uid);
  memcpy (&p, q, sizeof (struct passwd));

  return &p;
#else
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  static struct passwd p = { NULL, NULL, 0, 0, NULL, NULL, NULL };
  char query[256];
  MYSQL_RES *result;
  MYSQL_ROW row;

  if (p.pw_name != NULL)
  {
    /* cache the last uid queried */
    if (p.pw_uid == uid)
      return &p;
    free (p.pw_name);
    p.pw_name = NULL;
  }

  snprintf (query, sizeof (query),
            "select username from dspam_virtual_uids where uid = '%d'", uid);

  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return NULL;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL)
    return NULL;

  row = mysql_fetch_row (result);
  if (row == NULL)
  {
    mysql_free_result (result);
    return NULL;
  }

  if (row[0] == NULL)
  {
    mysql_free_result (result);
    return NULL;
  }

  p.pw_name = strdup (row[0]);
  p.pw_uid = uid;

  mysql_free_result (result);
  return &p;
#endif
}

void
_mysql_drv_query_error (const char *error, const char *query)
{
  FILE *file;
  time_t tm = time (NULL);
  char ct[128];
  char fn[MAX_FILENAME_LENGTH];

  LOG (LOG_WARNING, "query error: %s: see sql.errors for more details",
       error);

  snprintf (fn, sizeof (fn), "%s/sql.errors", USERDIR);

  snprintf (ct, sizeof (ct), "%s", ctime (&tm));
  chomp (ct);

  file = fopen (fn, "a");

  if (file == NULL)
  {
    file_error (ERROR_FILE_WRITE, fn, strerror (errno));
    return;
  }

  fprintf (file, "[%s] %d: %s: %s\n", ct, (int) getpid (), error, query);
  fclose (file);
  return;
}

#ifdef VIRTUAL_USERS
struct passwd *
_mysql_drv_setpwnam (DSPAM_CTX * CTX, const char *name)
{
  char query[256];
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;

  snprintf (query, sizeof (query),
            "insert into dspam_virtual_uids (uid, username) values(NULL, '%s')",
            name);

  /* we need to fail, to prevent a potential loop - even if it was inserted
   * by another process */
  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return NULL;
  }

  return _mysql_drv_getpwnam (CTX, name);

}
#endif

/* Neural network functions */

int _ds_get_node(DSPAM_CTX * CTX, char *user, struct _ds_neural_record *node) {
  char query[256];
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  MYSQL_RES *result;
  MYSQL_ROW row;
  uid_t n_uid;

  if (user == NULL) 
    n_uid = node->uid;
  else {
    p = _mysql_drv_getpwnam (CTX, user);
    if (p == NULL)
    {
      LOGDEBUG ("_mysql_drv_get_spamtotals: unable to _mysql_drv_getpwnam(%s)",
                user);
      return EINVAL;
    }
    n_uid = p->pw_uid;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL) 
  {
    LOGDEBUG ("_mysql_drv_get_spamtotals: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  snprintf(query, sizeof(query), "select total_correct, total_incorrect from "
    "dspam_neural_data where uid = %d and node = %d", p->pw_uid, n_uid);

  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return EUNKNOWN;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL)
    goto NEWNODE;

  row = mysql_fetch_row (result);
  if (row == NULL)
  {
    mysql_free_result (result);
    goto NEWNODE;
  }
  
  node->uid = n_uid;
  node->total_correct = strtol(row[0], NULL, 0);
  node->total_incorrect = strtol(row[1], NULL, 0);
  node->control_correct = node->total_correct;
  node->control_incorrect = node->total_incorrect;
  node->disk = 'Y';

  mysql_free_result (result);

  return 0;

NEWNODE:
  node->uid = n_uid;
  node->total_correct = 0;
  node->total_incorrect = 0;
  node->control_correct = 0;
  node->control_incorrect = 0;
  node->disk = 'N';
  return 0;
}

int _ds_set_node (DSPAM_CTX * CTX, char *user, struct _ds_neural_record *node) {
  char query[256];
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  uid_t n_uid;
                                                                                                                              
  if (user == NULL)
    n_uid = node->uid;
  else {
    p = _mysql_drv_getpwnam (CTX, user);
    if (p == NULL)
    {
      LOGDEBUG ("_mysql_drv_get_spamtotals: unable to _mysql_drv_getpwnam(%s)",
                user);
      return EINVAL;
    }
    n_uid = p->pw_uid;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_mysql_drv_get_spamtotals: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  if (node->disk == 'N') {
    snprintf(query, sizeof(query), "insert into dspam_neural_data "
      "(uid, node, total_correct, total_incorrect) values ("
      "%d, %d, %ld, %ld)", p->pw_uid, node->uid, node->total_correct, 
        node->total_incorrect);

    if (MYSQL_RUN_QUERY (s->dbh, query))
      node->disk = 'Y';
  }

  if (node->disk == 'Y') {
    snprintf(query, sizeof(query), "update dspam_neural_data set "
      "total_correct = total_correct + %ld, "
      "total_incorrect = total_incorrect + %ld "
      "where uid = %d and node = %d", 
        node->total_correct - node->control_correct,
        node->total_incorrect - node->control_incorrect,
      p->pw_uid, node->uid);

    if (MYSQL_RUN_QUERY (s->dbh, query))
    {
      _mysql_drv_query_error (mysql_error (s->dbh), query);
      return EUNKNOWN;
    }
  }

  node->disk = 'Y';
                                                                                
  return 0;

}

int
_ds_get_decision (DSPAM_CTX * CTX, struct _ds_neural_decision *DEC,
                  const char *signature)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  unsigned long *lengths;
  char *mem;
  char query[128];
  MYSQL_RES *result;
  MYSQL_ROW row;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_get_decision: invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_get_decision: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  snprintf (query, sizeof (query),
            "select data, length from dspam_neural_decisions where uid = %d and signature = \"%s\"",
            p->pw_uid, signature);
  if (mysql_real_query (s->dbh, query, strlen (query)))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return EFAILURE;
  }

  result = mysql_use_result (s->dbh);
  if (result == NULL)
    return EFAILURE;

  row = mysql_fetch_row (result);
  if (row == NULL)
  {
    mysql_free_result (result);
    return EFAILURE;
  }

  lengths = mysql_fetch_lengths (result);
  if (lengths == NULL || lengths[0] == 0)
  {
    mysql_free_result (result);
    return EFAILURE;
  }

  mem = malloc (lengths[0]);
  if (mem == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    mysql_free_result (result);
    return EUNKNOWN;
  }

  memcpy (mem, row[0], lengths[0]);
  DEC->data = mem;
  DEC->length = strtol (row[1], NULL, 0);

  mysql_free_result (result);

  return 0;
}

int
_ds_set_decision (DSPAM_CTX * CTX, struct _ds_neural_decision *DEC,
                   const char *signature)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  unsigned long length;
  char *mem;
  char scratch[1024];
  buffer *query;
  struct passwd *p;

  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_set_decision; invalid database handle (NULL)");
    return EINVAL;
  }

  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);

  if (p == NULL)
  {
    LOGDEBUG ("_ds_set_decision: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }

  query = buffer_create (NULL);
  if (query == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }

  mem = malloc (DEC->length * 2);
  if (mem == NULL)
  {
    LOG (LOG_CRIT, ERROR_MEM_ALLOC);
    return EUNKNOWN;
  }
  length = mysql_real_escape_string (s->dbh, mem, DEC->data, DEC->length);

  snprintf (scratch, sizeof (scratch),
            "insert into dspam_neural_decisions(uid, signature, length, created_on, data) values(%d, \"%s\", %ld, current_date(), \"",
            p->pw_uid, signature, DEC->length);
  buffer_cat (query, scratch);
  buffer_cat (query, mem);
  buffer_cat (query, "\")");

  if (mysql_real_query (s->dbh, query->data, query->used))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query->data);
    return EFAILURE;
  }

  free (mem);
  return 0;
}

int
_ds_delete_decision (DSPAM_CTX * CTX, const char *signature)
{
  struct _mysql_drv_storage *s = (struct _mysql_drv_storage *) CTX->storage;
  struct passwd *p;
  char query[128];
                                                                                                                              
  if (s->dbh == NULL)
  {
    LOGDEBUG ("_ds_delete_decision: invalid database handle (NULL)");
    return EINVAL;
  }
                                                                                                                              
  if (!CTX->group)
    p = _mysql_drv_getpwnam (CTX, CTX->username);
  else
    p = _mysql_drv_getpwnam (CTX, CTX->group);
                                                                                                                              
  if (p == NULL)
  {
    LOGDEBUG ("_ds_delete_decision: unable to _mysql_drv_getpwnam(%s)",
              CTX->username);
    return EINVAL;
  }
                                                                                                                              
  snprintf (query, sizeof (query),
            "delete from dspam_neural_decisions where uid = %d and signature = \"%s\"",
            p->pw_uid, signature);
  if (MYSQL_RUN_QUERY (s->dbh, query))
  {
    _mysql_drv_query_error (mysql_error (s->dbh), query);
    return EFAILURE;
  }
                                                                                                                              
  return 0;
}

