/*
 DSPAM
 COPYRIGHT (C) 2003 NETWORK DWEEBS CORPORATION

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

#ifdef HAVE_CONFIG_H
#include <auto-config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <dirent.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>

#include "../libdspam.h"

#define TSYNTAX	"syntax: dspam_stats [username]"

DSPAM_CTX *open_ctx, *open_mtx;

int stat_user (const char *username);
int process_all_users (void);
void dieout (int signal);

int
main (int argc, char **argv)
{

  open_ctx = open_mtx = NULL;

  signal (SIGINT, dieout);
  signal (SIGPIPE, dieout);
  signal (SIGTERM, dieout);

  dspam_init_driver ();

  if (argc > 1)
  {
    stat_user (argv[1]);
  }
  else
  {
    process_all_users ();
  }

  dspam_shutdown_driver ();
  exit (EXIT_SUCCESS);
}

int
process_all_users (void)
{
  DSPAM_CTX *CTX;
  char *user;

  CTX = dspam_init (NULL, NULL, DSM_TOOLS, 0);
  open_ctx = CTX;
  if (CTX == NULL)
  {
    fprintf (stderr, "Could not initialize context: %s\n", strerror (errno));
    exit (EXIT_FAILURE);
  }

  user = _ds_get_nextuser (CTX);
  while (user != NULL)
  {
    stat_user (user);
    user = _ds_get_nextuser (CTX);
  }

  dspam_destroy (CTX);
  open_ctx = NULL;
  return 0;
}

int
stat_user (const char *username)
{
  DSPAM_CTX *MTX;

  MTX = dspam_init (username, NULL, DSM_CLASSIFY, 0);
  open_mtx = MTX;
  if (MTX == NULL)
  {
    fprintf (stderr, "Could not init context: %s\n", strerror (errno));
    return EUNKNOWN;
  }

  printf ("%-16s  TS:% 6ld TI:% 6ld SM:% 6ld IM:% 6ld SC:% 6ld IC:% 6ld\n",
          username,
          MTX->totals.total_spam - (MTX->totals.spam_misclassified +
                                    MTX->totals.spam_corpusfed),
          MTX->totals.total_innocent - (MTX->totals.innocent_misclassified +
                                        MTX->totals.innocent_corpusfed),
          MTX->totals.spam_misclassified, MTX->totals.innocent_misclassified,
          MTX->totals.spam_corpusfed, MTX->totals.innocent_corpusfed);

  dspam_destroy (MTX);
  open_mtx = NULL;
  return 0;
}

void
dieout (int signal)
{
  fprintf (stderr, "terminated.\n");
  if (open_ctx != NULL)
    dspam_destroy (open_ctx);
  if (open_mtx != NULL)
    dspam_destroy (open_mtx);
  exit (EXIT_SUCCESS);
}
